import dataclasses
from znflow import utils
from znflow.base import NodeBaseMixin
import typing



@dataclasses.dataclass(frozen=True)
class Connection:
    """A Connector for Nodes.
    
    instance: either a Node or FunctionFuture
    attribute:
        Node.attribute
        or FunctionFuture.result
        or None if the class is passed and not an attribute
    """

    instance: any
    attribute: any
    item: any = None

    def __getitem__(self, item):
        return dataclasses.replace(self, instance=self, attribute="result", item=item)

    def __post_init__(self):
        if self.attribute is not None and self.attribute.startswith("_"):
            raise ValueError("Private attributes are not allowed.")

    @property
    def uuid(self):
        return self.instance.uuid

    @property
    def result(self):
        result = (
            getattr(self.instance, self.attribute) if self.attribute else self.instance
        )

        return result[self.item] if self.item else result
    
class UpdateConnectors(utils.IterableHandler):
    def default(self, value, **kwargs):
        return value.result if isinstance(value, Connection) else value
    
@dataclasses.dataclass
class FunctionFuture(NodeBaseMixin):
    function: typing.Callable
    args: typing.Tuple
    kwargs: typing.Dict
    item: any = None

    _result: any = dataclasses.field(default=None, init=False, repr=True)

    _protected_ = NodeBaseMixin._protected_ + ["function", "args", "kwargs"]

    def run(self):
        self._result = self.function(*self.args, **self.kwargs)

    def __getitem__(self, item):
        return Connection(instance=self, attribute="result", item=item)

    @property
    def result(self):
        if self._result is None:
            self.run()
        return self._result





"""Utils of the 'ZnFlow' package."""
import abc
import functools


class IterableHandler(abc.ABC):
    """Iterable handler using single dispatch.

    Attributes
    ----------
    updated : bool, default=False
        Set to True if handle updated data.
    """

    updated: bool = False

    def __init__(self):
        """Update the signature of handle based on the default method."""
        functools.update_wrapper(self.handle, self.default)
        self.updated = False

    def __call__(self, value, **kwargs):
        """Call the handle method."""
        return self.handle(value, **kwargs)

    @abc.abstractmethod
    def default(self, value, **kwargs):
        """How to handle the individual element inside the iterable.

        The first argument must be a positional only argument.
        All other arguments must be keyword only arguments.
        """
        raise NotImplementedError

    @functools.singledispatchmethod
    def handle(self, value, **kwargs):
        """Fallback handling if no siggledispatch was triggered."""

        result = self.default(value, **kwargs)
        if result is not value:
            self.updated = True
        return result

    @handle.register
    def _(self, value: list, **kwargs) -> list:
        """Handle a list."""
        return [self.handle(x, **kwargs) for x in value]

    @handle.register
    def _(self, value: tuple, **kwargs) -> tuple:
        """Handle a tuple."""
        # without 'tuple' it would be a generator
        return tuple(self.handle(x, **kwargs) for x in value)

    @handle.register
    def _(self, value: set, **kwargs) -> set:
        """Handle a set."""
        return {self.handle(x, **kwargs) for x in value}

    @handle.register
    def _(self, value: dict, **kwargs) -> dict:
        """Handle a dict."""
        return {key: self.handle(val, **kwargs) for key, val in value.items()}



from __future__ import annotations

import contextlib
import dataclasses
import typing
from uuid import UUID


@contextlib.contextmanager
def disable_graph(*args, **kwargs):
    """Temporarily disable set the graph to None.

    This can be useful, if you e.g. want to use 'get_attribute'.
    """
    graph = get_graph()
    set_graph(None)
    try:
        yield
    finally:
        set_graph(graph)


class Property:
    """Custom Property with disabled graph.

    References
    ----------
    Adapted from https://docs.python.org/3/howto/descriptor.html#properties
    """

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self.fget = disable_graph()(fget)
        self.fset = disable_graph()(fset)
        self.fdel = disable_graph()(fdel)
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc
        self._name = ""

    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        if self.fget is None:
            raise AttributeError(f"property '{self._name}' has no getter")
        return self.fget(obj)

    def __set__(self, obj, value):
        if self.fset is None:
            raise AttributeError(f"property '{self._name}' has no setter")
        self.fset(obj, value)

    def __delete__(self, obj):
        if self.fdel is None:
            raise AttributeError(f"property '{self._name}' has no deleter")
        self.fdel(obj)

    def getter(self, fget):
        prop = type(self)(fget, self.fset, self.fdel, self.__doc__)
        prop._name = self._name
        return prop

    def setter(self, fset):
        prop = type(self)(self.fget, fset, self.fdel, self.__doc__)
        prop._name = self._name
        return prop

    def deleter(self, fdel):
        prop = type(self)(self.fget, self.fset, fdel, self.__doc__)
        prop._name = self._name
        return prop


class NodeBaseMixin:
    """A Parent for all Nodes.

    This class is used to globally access and change all classes that inherit from it.

    Attributes
    ----------
    _graph_ : DiGraph
    uuid : UUID
    """

    _graph_ = None
    _uuid: UUID = None

    _protected_ = [
        "_graph_",
        "uuid",
        "_uuid",
        "result",
    ]  # TODO consider adding regex patterns

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        if self._uuid is not None:
            raise ValueError("uuid is already set")
        self._uuid = value

    def run(self):
        raise NotImplementedError


def get_graph():
    return NodeBaseMixin._graph_


def set_graph(value):
    NodeBaseMixin._graph_ = value


_get_attribute_none = object()


def get_attribute(obj, name, default=_get_attribute_none):
    """Get the real value of the attribute and not a znflow.Connection."""
    with disable_graph():
        if default is _get_attribute_none:
            return getattr(obj, name)
        return getattr(obj, name, default)


@dataclasses.dataclass(frozen=True)
class Connection:
    """A Connector for Nodes.
    instance: either a Node or FunctionFuture
    attribute:
        Node.attribute
        or FunctionFuture.result
        or None if the class is passed and not an attribute
    """

    instance: any
    attribute: any
    item: any = None

    def __getitem__(self, item):
        return dataclasses.replace(self, instance=self, attribute="result", item=item)

    def __post_init__(self):
        if self.attribute is not None and self.attribute.startswith("_"):
            raise ValueError("Private attributes are not allowed.")

    @property
    def uuid(self):
        return self.instance.uuid

    @property
    def result(self):
        result = (
            getattr(self.instance, self.attribute) if self.attribute else self.instance
        )

        return result[self.item] if self.item else result


@dataclasses.dataclass
class FunctionFuture(NodeBaseMixin):
    function: typing.Callable
    args: typing.Tuple
    kwargs: typing.Dict
    item: any = None

    _result: any = dataclasses.field(default=None, init=False, repr=True)

    _protected_ = NodeBaseMixin._protected_ + ["function", "args", "kwargs"]

    def run(self):
        self._result = self.function(*self.args, **self.kwargs)

    def __getitem__(self, item):
        return Connection(instance=self, attribute="result", item=item)

    @property
    def result(self):
        if self._result is None:
            self.run()
        return self._result



from __future__ import annotations

import functools
import inspect
import uuid

from znflow.base import (
    Connection,
    FunctionFuture,
    NodeBaseMixin,
    disable_graph,
    get_graph,
)


def _mark_init_in_construction(cls):
    if "__init__" in dir(cls):

        def wrap_init(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                cls._in_construction = True
                value = func(*args, **kwargs)
                cls._in_construction = False
                return value

            return wrapper

        cls.__init__ = wrap_init(cls.__init__)
    return cls


class Node(NodeBaseMixin):
    _in_construction = False

    def run(self):
        raise NotImplementedError

    def __matmul__(self, other):
        return Connection(self, other)

    def __new__(cls, *args, **kwargs):
        cls._in_construction = True
        try:
            instance = super().__new__(cls, *args, **kwargs)
        except TypeError:  # e.g. in dataclasses the arguments are passed to __new__
            # print("TypeError: ...")
            instance = super().__new__(cls)
        cls._in_construction = False
        _mark_init_in_construction(cls)
        instance.uuid = uuid.uuid4()

        # Connect the Node to the Grap
        graph = get_graph()
        if graph is not None:
            graph.add_node(instance)
        return instance

    def __getattribute__(self, item):
        if get_graph() is not None:
            with disable_graph():
                if item not in set(dir(self)):
                    raise AttributeError(
                        f"'{self.__class__.__name__}' object has no attribute '{item}'"
                    )

            if item not in type(self)._protected_ and not item.startswith("_"):
                if self._in_construction:
                    return super().__getattribute__(item)
                connector = Connection(instance=self, attribute=item)
                return connector
        return super().__getattribute__(item)

    def __setattr__(self, item, value) -> None:
        super().__setattr__(item, value)
        if get_graph() is not None:
            if isinstance(value, Connection):
                assert (
                    self.uuid in self._graph_
                ), f"'{self.uuid=}' not in '{self._graph_=}'"
                assert value.uuid in self._graph_
                self._graph_.add_edge(
                    value.uuid, self.uuid, u_attr=value.attribute, v_attr=item
                )


def nodify(function):
    """Decorator to create a Node from a function."""

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        """Wrapper function for the decorator.

        Raises
        ------
        TypeError:
            if the args / kwargs do not match the function signature
        """
        graph = get_graph()
        if graph is not None:
            # check if the args / kwargs match the function
            inspect.signature(function).bind(*args, **kwargs)

            future = FunctionFuture(function, args, kwargs)
            future.uuid = uuid.uuid4()

            graph.add_node(future)
            return future
        return function(*args, **kwargs)

    return wrapper



"""The 'ZnFlow' visualization module."""
import math

import networkx as nx


def get_count(node, graph) -> int:
    """Get the number of successors for the given node."""
    successors = list(graph.predecessors(node))
    return sum(get_count(x, graph) for x in successors) if len(successors) else 1


def get_colors(graph, log) -> list:
    """Get the color for each Node on the graph."""
    colors = [get_count(node, graph) for node in graph]
    return [math.log(x) for x in colors] if log else colors


def draw(graph, *args, log=True, **kwargs):
    """Draw the graph using networkx."""
    nx.draw(graph, *args, node_color=get_colors(graph, log), **kwargs)
